package com.sun.java.util.jar.pack;

import java.io.OutputStream;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.ArrayList;
import java.io.IOException;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.Deflater;
import java.util.Random;
import java.io.ByteArrayOutputStream;

class CodingChooser
{
    int verbose;
    int effort;
    boolean optUseHistogram;
    boolean optUsePopulationCoding;
    boolean optUseAdaptiveCoding;
    boolean disablePopCoding;
    boolean disableRunCoding;
    boolean topLevel;
    double fuzz;
    Coding[] allCodingChoices;
    CodingChooser.Choice[] choices;
    ByteArrayOutputStream context;
    CodingChooser popHelper;
    CodingChooser runHelper;
    Random stress;
    private int[] values;
    private int start;
    private int end;
    private int[] deltas;
    private int min;
    private int max;
    private Histogram vHist;
    private Histogram dHist;
    private int searchOrder;
    private CodingChooser.Choice regularChoice;
    private CodingChooser.Choice bestChoice;
    private CodingMethod bestMethod;
    private int bestByteSize;
    private int bestZipSize;
    private int targetSize;
    public static final int MIN_EFFORT = 1;
    public static final int MID_EFFORT = 5;
    public static final int MAX_EFFORT = 9;
    public static final int POP_EFFORT = 4;
    public static final int RUN_EFFORT = 3;
    public static final int BYTE_SIZE = 0;
    public static final int ZIP_SIZE = 1;
    private CodingChooser.Sizer zipSizer;
    private Deflater zipDef;
    private DeflaterOutputStream zipOut;
    private CodingChooser.Sizer byteSizer;
    private CodingChooser.Sizer byteOnlySizer;
    static final /* synthetic */ boolean $assertionsDisabled;
    
    CodingChooser(int effort, Coding[] allCodingChoices) {
        this.optUseHistogram = true;
        this.optUsePopulationCoding = true;
        this.optUseAdaptiveCoding = true;
        this.topLevel = true;
        this.zipSizer = new CodingChooser.Sizer();
        this.zipDef = new Deflater();
        this.zipOut = new DeflaterOutputStream(this.zipSizer, this.zipDef);
        this.byteSizer = new CodingChooser.Sizer(this.zipOut);
        this.byteOnlySizer = new CodingChooser.Sizer();
        PropMap currentPropMap = Utils.currentPropMap();
        if (currentPropMap != null) {
            this.verbose = Math.max(currentPropMap.getInteger("com.sun.java.util.jar.pack.verbose"), currentPropMap.getInteger("com.sun.java.util.jar.pack.verbose.coding"));
            this.optUseHistogram = !currentPropMap.getBoolean("com.sun.java.util.jar.pack.no.histogram");
            this.optUsePopulationCoding = !currentPropMap.getBoolean("com.sun.java.util.jar.pack.no.population.coding");
            this.optUseAdaptiveCoding = !currentPropMap.getBoolean("com.sun.java.util.jar.pack.no.adaptive.coding");
            int integer = currentPropMap.getInteger("com.sun.java.util.jar.pack.stress.coding");
            if (integer != 0) {
                this.stress = new Random((long)integer);
            }
        }
        this.effort = effort;
        this.allCodingChoices = allCodingChoices;
        this.fuzz = 1.0 + 0.0025 * (double)(effort - 5);
        int n = 0;
        for (int i = 0; i < allCodingChoices.length; ++i) {
            if (allCodingChoices[i] != null) {
                ++n;
            }
        }
        this.choices = new CodingChooser.Choice[n];
        int n2 = 0;
        for (int j = 0; j < allCodingChoices.length; ++j) {
            if (allCodingChoices[j] != null) {
                this.choices[n2++] = new CodingChooser.Choice(allCodingChoices[j], j, new int[this.choices.length]);
            }
        }
        int k = 0;
        while (k < this.choices.length) {
            Coding coding = this.choices[k].coding;
            assert coding.distanceFrom(coding) == 0;
            int l = 0;
            while (l < k) {
                Coding coding2 = this.choices[l].coding;
                int distance = coding.distanceFrom(coding2);
                assert distance > 0;
                assert distance == coding2.distanceFrom(coding);
                this.choices[k].distance[l] = distance;
                this.choices[l].distance[k] = distance;
                ++l;
            }
            ++k;
        }
    }
    
    CodingChooser.Choice makeExtraChoice(Coding coding) {
        int[] array = new int[this.choices.length];
        int i = 0;
        while (i < array.length) {
            Coding coding2 = this.choices[i].coding;
            int distance = coding.distanceFrom(coding2);
            assert distance > 0;
            assert distance == coding2.distanceFrom(coding);
            array[i] = distance;
            ++i;
        }
        CodingChooser.Choice o = new CodingChooser.Choice(coding, -1, array);
        o.reset();
        return o;
    }
    
    ByteArrayOutputStream getContext() {
        if (this.context == null) {
            this.context = new ByteArrayOutputStream(65536);
        }
        return this.context;
    }
    
    private void reset(int[] values, int start, int end) {
        this.values = values;
        this.start = start;
        this.end = end;
        this.deltas = null;
        this.min = Integer.MAX_VALUE;
        this.max = Integer.MIN_VALUE;
        this.vHist = null;
        this.dHist = null;
        this.searchOrder = 0;
        this.regularChoice = null;
        this.bestChoice = null;
        this.bestMethod = null;
        this.bestZipSize = Integer.MAX_VALUE;
        this.bestByteSize = Integer.MAX_VALUE;
        this.targetSize = Integer.MAX_VALUE;
    }
    
    CodingMethod choose(int[] array, int i, int n, Coding coding, int[] array2) {
        this.reset(array, i, n);
        if (this.effort <= 1 || i >= n) {
            if (array2 != null) {
                int[] computeSizePrivate = this.computeSizePrivate(coding);
                array2[0] = computeSizePrivate[0];
                array2[1] = computeSizePrivate[1];
                return coding;
            }
            return coding;
        }
        else {
            if (this.optUseHistogram) {
                this.getValueHistogram();
                this.getDeltaHistogram();
            }
            int n2;
            for (; i < n; ++i) {
                n2 = array[i];
                if (this.min > n2) {
                    this.min = n2;
                }
                if (this.max < n2) {
                    this.max = n2;
                }
            }
            int markUsableChoices = this.markUsableChoices(coding);
            if (this.stress != null) {
                int nextInt = this.stress.nextInt(markUsableChoices * 2 + 4);
                CodingMethod codingMethod = null;
                int k = 0;
                while (k < this.choices.length) {
                    CodingChooser.Choice choice = this.choices[k];
                    if (choice.searchOrder >= 0 && nextInt-- == 0) {
                        codingMethod = choice.coding;
                        break;
                    }
                    ++k;
                }
                if (codingMethod == null) {
                    if ((nextInt & 0x7) != 0x0) {
                        codingMethod = coding;
                    }
                    else {
                        codingMethod = this.stressCoding(this.min, this.max);
                    }
                }
                if (!this.disablePopCoding && this.optUsePopulationCoding && this.effort >= 4) {
                    codingMethod = this.stressPopCoding(codingMethod);
                }
                if (!this.disableRunCoding && this.optUseAdaptiveCoding && this.effort >= 3) {
                    codingMethod = this.stressAdaptiveCoding(codingMethod);
                    return codingMethod;
                }
                return codingMethod;
            }
            else {
                double n3 = 1.0;
                for (int j = this.effort; j < 9; ++j) {
                    n3 /= 1.414;
                }
                int n4 = (int)Math.ceil((double)markUsableChoices * n3);
                this.bestChoice = this.regularChoice;
                this.evaluate(this.regularChoice);
                int updateDistances = this.updateDistances(this.regularChoice);
                int bestZipSize = this.bestZipSize;
                int bestByteSize = this.bestByteSize;
                if (this.regularChoice.coding == coding && this.topLevel) {
                    int encodeEscapeValue = BandStructure.encodeEscapeValue(115, coding);
                    if (coding.canRepresentSigned(encodeEscapeValue)) {
                        int length = coding.getLength(encodeEscapeValue);
                        CodingChooser.Choice regularChoice = this.regularChoice;
                        regularChoice.zipSize -= length;
                        this.bestByteSize = this.regularChoice.byteSize;
                        this.bestZipSize = this.regularChoice.zipSize;
                    }
                }
                int n5 = 1;
                while (this.searchOrder < n4) {
                    if (n5 > updateDistances) {
                        n5 = 1;
                    }
                    CodingChooser.Choice choiceNear = this.findChoiceNear(this.bestChoice, updateDistances / n5, updateDistances / (n5 *= 2) + 1);
                    if (choiceNear == null) {
                        continue;
                    }
                    assert !(!choiceNear.coding.canRepresent(this.min, this.max));
                    this.evaluate(choiceNear);
                    int updateDistances2 = this.updateDistances(choiceNear);
                    if (choiceNear == this.bestChoice) {
                        updateDistances = updateDistances2;
                        if (this.verbose > 5) {
                            Utils.log.info("maxd = " + updateDistances);
                        }
                        else {
                            continue;
                        }
                    }
                    else {
                        continue;
                    }
                }
                Coding coding2 = this.bestChoice.coding;
                assert coding2 == this.bestMethod;
                if (this.verbose > 2) {
                    Utils.log.info("chooser: plain result=" + this.bestChoice + " after " + this.bestChoice.searchOrder + " rounds, " + (this.regularChoice.zipSize - this.bestZipSize) + " fewer bytes than regular " + coding);
                }
                this.bestChoice = null;
                if (!this.disablePopCoding && this.optUsePopulationCoding && this.effort >= 4 && this.bestMethod instanceof Coding) {
                    this.tryPopulationCoding(coding2);
                }
                if (!this.disableRunCoding && this.optUseAdaptiveCoding && this.effort >= 3 && this.bestMethod instanceof Coding) {
                    this.tryAdaptiveCoding(coding2);
                }
                if (array2 != null) {
                    array2[0] = this.bestByteSize;
                    array2[1] = this.bestZipSize;
                }
                if (this.verbose > 1) {
                    Utils.log.info("chooser: result=" + this.bestMethod + " " + (bestZipSize - this.bestZipSize) + " fewer bytes than regular " + coding + "; win=" + pct((double)(bestZipSize - this.bestZipSize), (double)bestZipSize));
                }
                CodingMethod bestMethod = this.bestMethod;
                this.reset(null, 0, 0);
                return bestMethod;
            }
        }
    }
    
    CodingMethod choose(int[] array, int n, int n2, Coding coding) {
        return this.choose(array, n, n2, coding, (int[])null);
    }
    
    CodingMethod choose(int[] array, Coding coding, int[] array2) {
        return this.choose(array, 0, array.length, coding, array2);
    }
    
    CodingMethod choose(int[] array, Coding coding) {
        return this.choose(array, 0, array.length, coding, (int[])null);
    }
    
    private int markUsableChoices(Coding coding) {
        int i = 0;
        for (int j = 0; j < this.choices.length; ++j) {
            CodingChooser.Choice regularChoice = this.choices[j];
            regularChoice.reset();
            if (!regularChoice.coding.canRepresent(this.min, this.max)) {
                regularChoice.searchOrder = -1;
                if (this.verbose > 1 && regularChoice.coding == coding) {
                    Utils.log.info("regular coding cannot represent [" + this.min + ".." + this.max + "]: " + coding);
                }
            }
            else {
                if (regularChoice.coding == coding) {
                    this.regularChoice = regularChoice;
                }
                ++i;
            }
        }
        if (this.regularChoice == null && coding.canRepresent(this.min, this.max)) {
            this.regularChoice = this.makeExtraChoice(coding);
            if (this.verbose > 1) {
                Utils.log.info("*** regular choice is extra: " + this.regularChoice.coding);
            }
        }
        if (this.regularChoice == null) {
            int k = 0;
            while (k < this.choices.length) {
                CodingChooser.Choice regularChoice2 = this.choices[k];
                if (regularChoice2.searchOrder != -1) {
                    this.regularChoice = regularChoice2;
                    break;
                }
                ++k;
            }
            if (this.verbose > 1) {
                Utils.log.info("*** regular choice does not apply " + coding);
                Utils.log.info("    using instead " + this.regularChoice.coding);
            }
        }
        if (this.verbose > 2) {
            Utils.log.info("chooser: #choices=" + i + " [" + this.min + ".." + this.max + "]");
            if (this.verbose > 4) {
                for (int l = 0; l < this.choices.length; ++l) {
                    CodingChooser.Choice obj = this.choices[l];
                    if (obj.searchOrder >= 0) {
                        Utils.log.info("  " + obj);
                    }
                }
            }
        }
        return i;
    }
    
    private CodingChooser.Choice findChoiceNear(CodingChooser.Choice obj, int i, int j) {
        if (this.verbose > 5) {
            Utils.log.info("findChoice " + i + ".." + j + " near: " + obj);
        }
        int[] distance = obj.distance;
        CodingChooser.Choice obj2 = null;
        for (int k = 0; k < this.choices.length; ++k) {
            CodingChooser.Choice obj3 = this.choices[k];
            if (obj3.searchOrder >= this.searchOrder && (distance[k] >= j && distance[k] <= i)) {
                if (obj3.minDistance >= j && obj3.minDistance <= i) {
                    if (this.verbose > 5) {
                        Utils.log.info("findChoice => good " + obj3);
                        return obj3;
                    }
                    return obj3;
                }
                else {
                    obj2 = obj3;
                }
            }
        }
        if (this.verbose > 5) {
            Utils.log.info("findChoice => found " + obj2);
            return obj2;
        }
        return obj2;
    }
    
    private void evaluate(CodingChooser.Choice choice) {
        assert choice.searchOrder == Integer.MAX_VALUE;
        choice.searchOrder = this.searchOrder++;
        int n;
        if (choice == this.bestChoice || choice.isExtra()) {
            n = 1;
        }
        else if (this.optUseHistogram) {
            choice.histSize = (int)Math.ceil(this.getHistogram(choice.coding.isDelta()).getBitLength(choice.coding) / 8.0);
            choice.byteSize = choice.histSize;
            n = ((choice.byteSize <= this.targetSize) ? 1 : 0);
        }
        else {
            n = 1;
        }
        if (n != 0) {
            int[] computeSizePrivate = this.computeSizePrivate(choice.coding);
            choice.byteSize = computeSizePrivate[0];
            choice.zipSize = computeSizePrivate[1];
            if (this.noteSizes(choice.coding, choice.byteSize, choice.zipSize)) {
                this.bestChoice = choice;
            }
        }
        if (choice.histSize >= 0 && !CodingChooser.$assertionsDisabled && choice.byteSize != choice.histSize) {
            throw new AssertionError();
        }
        if (this.verbose > 4) {
            Utils.log.info("evaluated " + choice);
        }
        return;
    }
    
    private boolean noteSizes(CodingMethod codingMethod, int n, int n2) {
        assert n2 > 0 && n > 0;
        int n3 = n2 < this.bestZipSize;
        if (this.verbose > 3) {
            Utils.log.info("computed size " + codingMethod + " " + n + "/zs=" + n2 + ((n3 && this.bestMethod != null) ? (" better by " + pct((double)(this.bestZipSize - n2), (double)n2)) : ""));
        }
        if (!(n3 == 0)) {
            this.bestMethod = codingMethod;
            this.bestZipSize = n2;
            this.bestByteSize = n;
            this.targetSize = (int)((double)n * this.fuzz);
            return true;
        }
        return false;
    }
    
    private int updateDistances(CodingChooser.Choice choice) {
        int[] distance = choice.distance;
        int i = 0;
        for (int j = 0; j < this.choices.length; ++j) {
            CodingChooser.Choice obj = this.choices[j];
            if (obj.searchOrder >= this.searchOrder) {
                int n = distance[j];
                if (this.verbose > 5) {
                    Utils.log.info("evaluate dist " + n + " to " + obj);
                }
                if (obj.minDistance > n) {
                    obj.minDistance = n;
                }
                if (i < n) {
                    i = n;
                }
            }
        }
        if (this.verbose > 5) {
            Utils.log.info("evaluate maxd => " + i);
        }
        return i;
    }
    
    public void computeSize(CodingMethod codingMethod, int[] array, int n, int n2, int[] array2) {
        if (n2 <= n) {
            array2[0] = (array2[1] = 0);
            return;
        }
        try {
            this.resetData();
            codingMethod.writeArrayTo(this.byteSizer, array, n, n2);
            array2[0] = this.getByteSize();
            array2[1] = this.getZipSize();
        }
        catch (IOException cause) {
            throw new RuntimeException(cause);
        }
        return;
    }
    
    public void computeSize(CodingMethod codingMethod, int[] array, int[] array2) {
        this.computeSize(codingMethod, array, 0, array.length, array2);
    }
    
    public int[] computeSize(CodingMethod codingMethod, int[] array, int n, int n2) {
        int[] array2 = { 0, 0 };
        this.computeSize(codingMethod, array, n, n2, array2);
        return array2;
    }
    
    public int[] computeSize(CodingMethod codingMethod, int[] array) {
        return this.computeSize(codingMethod, array, 0, array.length);
    }
    
    private int[] computeSizePrivate(CodingMethod codingMethod) {
        int[] array = { 0, 0 };
        this.computeSize(codingMethod, this.values, this.start, this.end, array);
        return array;
    }
    
    public int computeByteSize(CodingMethod obj, int[] array, int n, int n2) {
        if (n2 - n < 0) {
            return 0;
        }
        if ((obj instanceof Coding)) {
            int length = ((Coding)obj).getLength(array, n, n2);
            int countBytesToSizer;
            assert length == (countBytesToSizer = this.countBytesToSizer(obj, array, n, n2)) : obj + " : " + length + " != " + countBytesToSizer;
            return length;
        }
        else {
            return this.countBytesToSizer(obj, array, n, n2);
        }
    }
    
    private int countBytesToSizer(CodingMethod codingMethod, int[] array, int n, int n2) {
        try {
            this.byteOnlySizer.reset();
            codingMethod.writeArrayTo(this.byteOnlySizer, array, n, n2);
            return this.byteOnlySizer.getSize();
        }
        catch (IOException cause) {
            throw new RuntimeException(cause);
        }
    }
    
    int[] getDeltas(int n, int n2) {
        if ((n | n2) != 0x0) {
            return Coding.makeDeltas(this.values, this.start, this.end, n, n2);
        }
        if (this.deltas == null) {
            this.deltas = Coding.makeDeltas(this.values, this.start, this.end, 0, 0);
        }
        return this.deltas;
    }
    
    Histogram getValueHistogram() {
        if (this.vHist == null) {
            this.vHist = new Histogram(this.values, this.start, this.end);
            if (this.verbose > 3) {
                this.vHist.print("vHist", System.out);
            }
            else if (this.verbose > 1) {
                this.vHist.print("vHist", null, System.out);
            }
        }
        return this.vHist;
    }
    
    Histogram getDeltaHistogram() {
        if (this.dHist == null) {
            this.dHist = new Histogram(this.getDeltas(0, 0));
            if (this.verbose > 3) {
                this.dHist.print("dHist", System.out);
            }
            else if (this.verbose > 1) {
                this.dHist.print("dHist", null, System.out);
            }
        }
        return this.dHist;
    }
    
    Histogram getHistogram(boolean b) {
        return b ? this.getDeltaHistogram() : this.getValueHistogram();
    }
    
    private void tryPopulationCoding(Coding coding) {
        Histogram valueHistogram = this.getValueHistogram();
        Coding valueCoding = coding.getValueCoding();
        Coding setL = BandStructure.UNSIGNED5.setL(64);
        Coding valueCoding2 = coding.getValueCoding();
        int n = 4 + Math.max(valueCoding.getLength(this.min), valueCoding.getLength(this.max));
        int length = setL.getLength(0);
        int n2 = length * (this.end - this.start);
        int n3 = (int)Math.ceil(valueHistogram.getBitLength(valueCoding2) / 8.0);
        int i = n + n2 + n3;
        int min = 0;
        int[] array = new int[1 + valueHistogram.getTotalLength()];
        int j = -1;
        int n4 = -1;
        int[][] matrix = valueHistogram.getMatrix();
        int n5 = -1;
        int length2 = 1;
        int n6 = 0;
        for (int k = 1; k <= valueHistogram.getTotalLength(); ++k) {
            if (length2 == 1) {
                ++n5;
                n6 = matrix[n5][0];
                length2 = matrix[n5].length;
            }
            int n7 = matrix[n5][--length2];
            array[k] = n7;
            int length3 = valueCoding.getLength(n7);
            n += length3;
            n2 += (setL.getLength(k) - length) * n6;
            n3 -= length3 * n6;
            int n8 = n + n2 + n3;
            if (i > n8) {
                if (n8 <= this.targetSize) {
                    n4 = k;
                    if (j < 0) {
                        j = k;
                    }
                    if (this.verbose > 4) {
                        Utils.log.info("better pop-size at fvc=" + k + " by " + pct((double)(i - n8), (double)i));
                    }
                }
                i = n8;
                min = k;
            }
        }
        if (j < 0) {
            if (this.verbose > 1 && this.verbose > 1) {
                Utils.log.info("no good pop-size; best was " + i + " at " + min + " worse by " + pct((double)(i - this.bestByteSize), (double)this.bestByteSize));
            }
            return;
        }
        else {
            if (this.verbose > 1) {
                Utils.log.info("initial best pop-size at fvc=" + min + " in [" + j + ".." + n4 + "] by " + pct((double)(this.bestByteSize - i), (double)this.bestByteSize));
            }
            int bestZipSize = this.bestZipSize;
            int[] lValuesCoded = PopulationCoding.LValuesCoded;
            ArrayList o = new ArrayList();
            ArrayList o2 = new ArrayList();
            ArrayList o3 = new ArrayList();
            if (min <= 255) {
                o.add(BandStructure.BYTE1);
            }
            else {
                int b = 5;
                int n9 = this.effort > 4;
                if (!(n9 == 0)) {
                    o2.add(BandStructure.BYTE1.setS(1));
                }
                for (int l = lValuesCoded.length - 1; l >= 1; --l) {
                    int n10 = lValuesCoded[l];
                    Coding fitTokenCoding = PopulationCoding.fitTokenCoding(j, n10);
                    Coding fitTokenCoding2 = PopulationCoding.fitTokenCoding(min, n10);
                    Coding fitTokenCoding3 = PopulationCoding.fitTokenCoding(n4, n10);
                    if (fitTokenCoding2 != null) {
                        if (!o.contains(fitTokenCoding2)) {
                            o.add(fitTokenCoding2);
                        }
                        if (b > fitTokenCoding2.B()) {
                            b = fitTokenCoding2.B();
                        }
                    }
                    if (!(n9 == 0)) {
                        if (fitTokenCoding3 == null) {
                            fitTokenCoding3 = fitTokenCoding2;
                        }
                        for (int b2 = fitTokenCoding.B(); b2 <= fitTokenCoding3.B(); ++b2) {
                            if (b2 != fitTokenCoding2.B() && b2 != 1) {
                                Coding setS = fitTokenCoding3.setB(b2).setS(1);
                                if (!o2.contains(setS)) {
                                    o2.add(setS);
                                }
                            }
                        }
                    }
                }
                Iterator<Object> iterator = o.iterator();
                while (iterator.hasNext()) {
                    Coding coding2 = (Coding)iterator.next();
                    if (coding2.B() > b) {
                        iterator.remove();
                        o3.add(0, coding2);
                    }
                    else {
                        continue;
                    }
                }
            }
            ArrayList o4 = new ArrayList();
            Iterator<Object> iterator2 = o.iterator();
            Iterator<Object> iterator3 = o2.iterator();
            Iterator<Object> iterator4 = o3.iterator();
            while (iterator2.hasNext() || iterator3.hasNext() || iterator4.hasNext()) {
                if (iterator2.hasNext()) {
                    o4.add(iterator2.next());
                }
                if (iterator3.hasNext()) {
                    o4.add(iterator3.next());
                }
                if (!iterator4.hasNext()) {
                    continue;
                }
                o4.add(iterator4.next());
            }
            o.clear();
            o2.clear();
            o3.clear();
            int size = o4.size();
            if (this.effort == 4) {
                size = 2;
            }
            else if (size > 4) {
                size -= 4;
                size = size * (this.effort - 4) / 5;
                size += 4;
            }
            if (o4.size() > size) {
                if (this.verbose > 4) {
                    Utils.log.info("allFits before clip: " + o4);
                }
                o4.subList(size, o4.size()).clear();
            }
            if (this.verbose > 3) {
                Utils.log.info("allFits: " + o4);
            }
            for (Coding setS2 : o4) {
                boolean n11 = false;
                if (setS2.S() == 1) {
                    n11 = true;
                    setS2 = setS2.setS(0);
                }
                if (n11 == 0) {
                    assert setS2.umax() >= min;
                    assert setS2.setB(setS2.B() - 1).umax() < min;
                }
                else {
                    min = Math.min(setS2.umax(), n4);
                    if (min < j) {
                        continue;
                    }
                    if (min == min) {
                        continue;
                    }
                }
                PopulationCoding o5 = new PopulationCoding();
                o5.setHistogram(valueHistogram);
                o5.setL(setS2.L());
                o5.setFavoredValues(array, min);
                assert o5.tokenCoding == setS2;
                o5.resortFavoredValues();
                int[] computePopSizePrivate = this.computePopSizePrivate(o5, valueCoding, valueCoding2);
                this.noteSizes(o5, computePopSizePrivate[0], 4 + computePopSizePrivate[1]);
            }
            if (this.verbose > 3) {
                Utils.log.info("measured best pop, size=" + this.bestByteSize + "/zs=" + this.bestZipSize + " better by " + pct((double)(bestZipSize - this.bestZipSize), (double)bestZipSize));
                if (this.bestZipSize < bestZipSize) {
                    Utils.log.info(">>> POP WINS BY " + (bestZipSize - this.bestZipSize));
                }
            }
            return;
        }
    }
    
    private int[] computePopSizePrivate(PopulationCoding populationCoding, Coding obj, Coding coding) {
        if (this.popHelper == null) {
            this.popHelper = new CodingChooser(this.effort, this.allCodingChoices);
            if (this.stress != null) {
                this.popHelper.addStressSeed(this.stress.nextInt());
            }
            this.popHelper.topLevel = false;
            CodingChooser popHelper = this.popHelper;
            --popHelper.verbose;
            this.popHelper.disablePopCoding = true;
            this.popHelper.disableRunCoding = this.disableRunCoding;
            if (this.effort < 5) {
                this.popHelper.disableRunCoding = true;
            }
        }
        int fVlen = populationCoding.fVlen;
        if (this.verbose > 2) {
            Utils.log.info("computePopSizePrivate fvlen=" + fVlen + " tc=" + populationCoding.tokenCoding);
            Utils.log.info("{ //BEGIN");
        }
        int[] fValues = populationCoding.fValues;
        int[][] encodeValues = populationCoding.encodeValues(this.values, this.start, this.end);
        int[] array = encodeValues[0];
        int[] array2 = encodeValues[1];
        if (this.verbose > 2) {
            Utils.log.info("-- refine on fv[" + fVlen + "] fc=" + obj);
        }
        populationCoding.setFavoredCoding(this.popHelper.choose(fValues, 1, 1 + fVlen, obj));
        if (populationCoding.tokenCoding instanceof Coding && (this.stress == null || this.stress.nextBoolean())) {
            if (this.verbose > 2) {
                Utils.log.info("-- refine on tv[" + array.length + "] tc=" + populationCoding.tokenCoding);
            }
            CodingMethod choose = this.popHelper.choose(array, (Coding)populationCoding.tokenCoding);
            if (choose != populationCoding.tokenCoding) {
                if (this.verbose > 2) {
                    Utils.log.info(">>> refined tc=" + choose);
                }
                populationCoding.setTokenCoding(choose);
            }
        }
        if (array2.length == 0) {
            populationCoding.setUnfavoredCoding(null);
        }
        else {
            if (this.verbose > 2) {
                Utils.log.info("-- refine on uv[" + array2.length + "] uc=" + populationCoding.unfavoredCoding);
            }
            populationCoding.setUnfavoredCoding(this.popHelper.choose(array2, coding));
        }
        if (this.verbose > 3) {
            Utils.log.info("finish computePopSizePrivate fvlen=" + fVlen + " fc=" + populationCoding.favoredCoding + " tc=" + populationCoding.tokenCoding + " uc=" + populationCoding.unfavoredCoding);
            StringBuilder o = new StringBuilder();
            o.append("fv = {");
            for (int i = 1; i <= fVlen; ++i) {
                if (i % 10 == 0) {
                    o.append('\n');
                }
                o.append(" ").append(fValues[i]);
            }
            o.append('\n');
            o.append("}");
            Utils.log.info(o.toString());
        }
        if (this.verbose > 2) {
            Utils.log.info("} //END");
        }
        if (this.stress != null) {
            return null;
        }
        int[] array3;
        try {
            this.resetData();
            populationCoding.writeSequencesTo(this.byteSizer, array, array2);
            array3 = new int[] { this.getByteSize(), this.getZipSize() };
        }
        catch (IOException cause) {
            throw new RuntimeException(cause);
        }
        int[] computeSizePrivate = null;
        assert (computeSizePrivate = this.computeSizePrivate(populationCoding)) != null;
        assert computeSizePrivate[0] == array3[0] : computeSizePrivate[0] + " != " + array3[0];
        return array3;
    }
    
    private void tryAdaptiveCoding(Coding codingMethod) {
        int bestZipSize = this.bestZipSize;
        int i = this.start;
        int n = this.end;
        int[] array = this.values;
        int j = n - i;
        if (codingMethod.isDelta()) {
            array = this.getDeltas(0, 0);
            i = 0;
            n = array.length;
        }
        int[] array2 = new int[j + 1];
        int n2 = 0;
        int n3 = 0;
        while (i < n) {
            int n4 = array[i];
            array2[n2++] = n3;
            int length = codingMethod.getLength(n4);
            assert length < Integer.MAX_VALUE;
            n3 += length;
            ++i;
        }
        array2[n2++] = n3;
        assert n2 == array2.length;
        double d = (double)n3 / (double)j;
        double n5;
        if (this.effort >= 5) {
            if (this.effort > 6) {
                n5 = 1.001;
            }
            else {
                n5 = 1.003;
            }
        }
        else if (this.effort > 3) {
            n5 = 1.01;
        }
        else {
            n5 = 1.03;
        }
        double d2 = n5 * n5;
        double n6 = d2 * d2;
        double n7 = d2 * d2 * d2;
        double[] array3 = new double[1 + (this.effort - 3)];
        double log = Math.log((double)j);
        for (int k = 0; k < array3.length; ++k) {
            array3[k] = Math.exp(log * (double)(k + 1) / (double)(array3.length + 1));
        }
        int[] array4 = new int[array3.length];
        int n8 = 0;
        for (int l = 0; l < array3.length; ++l) {
            int nextK = AdaptiveCoding.getNextK((int)Math.round(array3[l]) - 1);
            if (nextK > 0 && nextK < j && (n8 <= 0 || nextK != array4[n8 - 1])) {
                array4[n8++] = nextK;
            }
        }
        int[] realloc = BandStructure.realloc(array4, n8);
        int[] array5 = new int[realloc.length];
        double[] array6 = new double[realloc.length];
        int n10;
        for (int n9 = 0; n9 < realloc.length; array6[n9] = n7, array5[n9] = 4 + (int)Math.ceil((double)n10 * d * n7), ++n9) {
            n10 = realloc[n9];
            if (n10 < 10) {
            }
            else if (n10 < 100) {
                n7 = n6;
            }
            else {
                n7 = d2;
            }
        }
        if (this.verbose > 1) {
            System.out.print("tryAdaptiveCoding [" + j + "] avgS=" + d + " fuzz=" + d2 + " meshes: {");
            for (int n11 = 0; n11 < realloc.length; ++n11) {
                System.out.print(" " + realloc[n11] + "(" + array5[n11] + ")");
            }
            Utils.log.info(" }");
        }
        if (this.runHelper == null) {
            this.runHelper = new CodingChooser(this.effort, this.allCodingChoices);
            if (this.stress != null) {
                this.runHelper.addStressSeed(this.stress.nextInt());
            }
            this.runHelper.topLevel = false;
            CodingChooser runHelper = this.runHelper;
            --runHelper.verbose;
            this.runHelper.disableRunCoding = true;
            this.runHelper.disablePopCoding = this.disablePopCoding;
            if (this.effort < 5) {
                this.runHelper.disablePopCoding = true;
            }
        }
        for (int nextK2 = 0; nextK2 < j; ++nextK2) {
            nextK2 = AdaptiveCoding.getNextK(nextK2 - 1);
            if (nextK2 > j) {
                nextK2 = j;
            }
            for (int n12 = realloc.length - 1; n12 >= 0; --n12) {
                int n13 = realloc[n12];
                int n14 = array5[n12];
                if (nextK2 + n13 <= j) {
                    int n15 = array2[nextK2 + n13] - array2[nextK2];
                    if (n15 >= n14) {
                        j = nextK2 + n13;
                        int n16 = n15;
                        double n17 = d * array6[n12];
                        while (j < j && j - nextK2 <= j / 2) {
                            j = nextK2 + AdaptiveCoding.getNextK(j + n13 - nextK2 - 1);
                            if (j < 0 || j > j) {
                            }
                            n16 = array2[j] - array2[nextK2];
                            if ((double)n16 < 4.0 + (double)(j - nextK2) * n17) {
                                break;
                            }
                            continue;
                        }
                        if (this.verbose > 2) {
                            Utils.log.info("bulge at " + nextK2 + "[" + (j - nextK2) + "] of " + pct((double)n16 - d * (double)(j - nextK2), d * (double)(j - nextK2)));
                            Utils.log.info("{ //BEGIN");
                        }
                        CodingMethod choose = this.runHelper.choose(this.values, this.start + nextK2, this.start + j, codingMethod);
                        if (choose == codingMethod) {
                            codingMethod = (codingMethod = codingMethod);
                        }
                        else {
                            codingMethod = this.runHelper.choose(this.values, this.start, this.start + nextK2, codingMethod);
                            codingMethod = this.runHelper.choose(this.values, this.start + j, this.start + j, codingMethod);
                        }
                        if (this.verbose > 2) {
                            Utils.log.info("} //END");
                        }
                        if (codingMethod == choose && nextK2 > 0 && AdaptiveCoding.isCodableLength(j)) {
                            nextK2 = 0;
                        }
                        if (choose == codingMethod && j < j) {
                        }
                        if (codingMethod != codingMethod || choose != codingMethod || codingMethod != codingMethod) {
                            int n18 = 0;
                            Object o;
                            if (j == j) {
                                o = choose;
                            }
                            else {
                                o = new AdaptiveCoding(j - nextK2, choose, codingMethod);
                                n18 += 4;
                            }
                            if (nextK2 > 0) {
                                o = new AdaptiveCoding(nextK2, codingMethod, o);
                                n18 += 4;
                            }
                            int[] computeSizePrivate = this.computeSizePrivate(o);
                            this.noteSizes(o, computeSizePrivate[0], computeSizePrivate[1] + n18);
                        }
                        nextK2 = j;
                        break;
                    }
                }
            }
        }
        if (this.verbose > 3 && this.bestZipSize < bestZipSize) {
            Utils.log.info(">>> RUN WINS BY " + (bestZipSize - this.bestZipSize));
        }
        return;
    }
    
    private static String pct(double n, double n2) {
        return (double)Math.round(n / n2 * 10000.0) / 100.0 + "%";
    }
    
    private void resetData() {
        this.flushData();
        this.zipDef.reset();
        if (this.context != null) {
            try {
                this.context.writeTo(this.byteSizer);
            }
            catch (IOException cause) {
                throw new RuntimeException(cause);
            }
        }
        this.zipSizer.reset();
        this.byteSizer.reset();
    }
    
    private void flushData() {
        try {
            this.zipOut.finish();
        }
        catch (IOException cause) {
            throw new RuntimeException(cause);
        }
    }
    
    private int getByteSize() {
        return this.byteSizer.getSize();
    }
    
    private int getZipSize() {
        this.flushData();
        return this.zipSizer.getSize();
    }
    
    void addStressSeed(int n) {
        if (this.stress == null) {
            return;
        }
        this.stress.setSeed((long)n + ((long)this.stress.nextInt() << 32));
        return;
    }
    
    private CodingMethod stressPopCoding(CodingMethod codingMethod) {
        assert this.stress != null;
        if (!(codingMethod instanceof Coding)) {
            return codingMethod;
        }
        Coding valueCoding = ((Coding)codingMethod).getValueCoding();
        Histogram valueHistogram = this.getValueHistogram();
        int stressLen = this.stressLen(valueHistogram.getTotalLength());
        if (stressLen == 0) {
            return codingMethod;
        }
        ArrayList list = new ArrayList();
        if (this.stress.nextBoolean()) {
            HashSet<Integer> set = new HashSet();
            for (int k = this.start; k < this.end; ++k) {
                if (set.add(Integer.valueOf(this.values[k]))) {
                    list.add(Integer.valueOf(this.values[k]));
                }
            }
        }
        else {
            int[][] matrix = valueHistogram.getMatrix();
            for (int i = 0; i < matrix.length; ++i) {
                int[] array = matrix[i];
                for (int j = 1; j < array.length; ++j) {
                    list.add(Integer.valueOf(array[j]));
                }
            }
        }
        int nextInt = this.stress.nextInt();
        if ((nextInt & 0x7) <= 2) {
            Collections.shuffle(list, this.stress);
        }
        else {
            int n;
            if (((n = nextInt >>> 3) & 0x7) <= 2) {
                Collections.sort(list);
            }
            int n2;
            if (((n2 = n >>> 3) & 0x7) <= 2) {
                Collections.reverse(list);
            }
            if (((nextInt = n2 >>> 3) & 0x7) <= 2) {
                Collections.rotate(list, this.stressLen(list.size()));
            }
        }
        if (list.size() > stressLen) {
            if ((nextInt >>> 3 & 0x7) <= 2) {
                list.subList(stressLen, list.size()).clear();
            }
            else {
                list.subList(0, list.size() - stressLen).clear();
            }
        }
        int size = list.size();
        int[] array2 = new int[1 + size];
        for (int l = 0; l < size; ++l) {
            array2[1 + l] = (int)list.get(l);
        }
        PopulationCoding o = new PopulationCoding();
        o.setFavoredValues(array2, size);
        int[] lValuesCoded = PopulationCoding.LValuesCoded;
        for (int n3 = 0; n3 < lValuesCoded.length / 2; ++n3) {
            int m = lValuesCoded[this.stress.nextInt(lValuesCoded.length)];
            if (m >= 0 && PopulationCoding.fitTokenCoding(size, m) != null) {
                o.setL(m);
                break;
            }
        }
        if (o.tokenCoding == null) {
            int n4 = array2[1];
            for (int n5 = 2; n5 <= size; ++n5) {
                int n6 = array2[n5];
                if (n4 > n6) {
                    n4 = n6;
                }
                if (n4 < n6) {
                    n4 = n6;
                }
            }
            o.tokenCoding = this.stressCoding(n4, n4);
        }
        this.computePopSizePrivate(o, valueCoding, valueCoding);
        return o;
    }
    
    private CodingMethod stressAdaptiveCoding(CodingMethod codingMethod) {
        assert this.stress != null;
        if (!(codingMethod instanceof Coding)) {
            return codingMethod;
        }
        Coding coding = (Coding)codingMethod;
        int n = this.end - this.start;
        if (n < 2) {
            return codingMethod;
        }
        int n2 = this.stressLen(n - 1) + 1;
        if (n2 == n) {
            return codingMethod;
        }
        try {
            assert !this.disableRunCoding;
            this.disableRunCoding = true;
            int[] array = (int[])this.values.clone();
            Object o = null;
            int i = this.end;
            int start = this.start;
            while (i > start) {
                int n3 = (i - start < 100) ? -1 : this.stress.nextInt();
                int decodeK;
                if ((n3 & 0x7) != 0x0) {
                    decodeK = ((n2 == 1) ? n2 : (this.stressLen(n2 - 1) + 1));
                }
                else {
                    int n5;
                    int n4 = (n5 = n3 >>> 3) & 0x3;
                    int n6 = n5 >>> 3 & 0xFF;
                    while (true) {
                        decodeK = AdaptiveCoding.decodeK(n4, n6);
                        if (decodeK <= i - start) {
                            break;
                        }
                        if (n6 != 3) {
                            n6 = 3;
                        }
                        else {
                            --n4;
                        }
                    }
                    assert !(!AdaptiveCoding.isCodableLength(decodeK));
                }
                if (decodeK > i - start) {
                    decodeK = i - start;
                }
                while (!AdaptiveCoding.isCodableLength(decodeK)) {
                    --decodeK;
                }
                int n7 = i - decodeK;
                assert n7 < i;
                assert n7 >= start;
                CodingMethod choose = this.choose(array, n7, i, coding);
                if (o == null) {
                    o = choose;
                }
                else {
                    o = new AdaptiveCoding(i - n7, choose, o);
                }
                i = n7;
            }
            return o;
        }
        finally {
            this.disableRunCoding = false;
        }
    }
    
    private Coding stressCoding(int n, int n2) {
        assert this.stress != null;
        int i = 0;
        while (i < 100) {
            Coding coding = Coding.of(this.stress.nextInt(5) + 1, this.stress.nextInt(256) + 1, this.stress.nextInt(3));
            if (coding.B() == 1) {
                coding = coding.setH(256);
            }
            if (coding.H() == 256 && coding.B() >= 5) {
                coding = coding.setB(4);
            }
            if (this.stress.nextBoolean()) {
                Coding setD = coding.setD(1);
                if (setD.canRepresent(n, n2)) {
                    return setD;
                }
            }
            if (coding.canRepresent(n, n2)) {
                return coding;
            }
            ++i;
        }
        return BandStructure.UNSIGNED5;
    }
    
    private int stressLen(int bound) {
        assert this.stress != null;
        assert bound >= 0;
        int nextInt = this.stress.nextInt(100);
        if (nextInt < 20) {
            return Math.min(bound / 5, nextInt);
        }
        if (nextInt < 40) {
            return bound;
        }
        return this.stress.nextInt(bound);
    }
    
    static class Choice
    {
        final Coding coding;
        final int index;
        final int[] distance;
        int searchOrder;
        int minDistance;
        int zipSize;
        int byteSize;
        int histSize;
        
        Choice(Coding coding, int index, int[] distance) {
            this.coding = coding;
            this.index = index;
            this.distance = distance;
        }
        
        void reset() {
            this.searchOrder = Integer.MAX_VALUE;
            this.minDistance = Integer.MAX_VALUE;
            int zipSize = -1;
            this.histSize = zipSize;
            this.byteSize = zipSize;
            this.zipSize = zipSize;
        }
        
        boolean isExtra() {
            return this.index < 0;
        }
        
        @Override
        public String toString() {
            return this.stringForDebug();
        }
        
        private String stringForDebug() {
            String s = "";
            if (this.searchOrder < Integer.MAX_VALUE) {
                s = s + " so: " + this.searchOrder;
            }
            if (this.minDistance < Integer.MAX_VALUE) {
                s = s + " md: " + this.minDistance;
            }
            if (this.zipSize > 0) {
                s = s + " zs: " + this.zipSize;
            }
            if (this.byteSize > 0) {
                s = s + " bs: " + this.byteSize;
            }
            if (this.histSize > 0) {
                s = s + " hs: " + this.histSize;
            }
            return "Choice[" + this.index + "] " + s + " " + this.coding;
        }
    }
    
    static class Sizer extends OutputStream
    {
        final OutputStream out;
        private int count;
        
        Sizer(OutputStream out) {
            this.out = out;
        }
        
        Sizer() {
            this(null);
        }
        
        @Override
        public void write(int n) throws IOException {
            ++this.count;
            if (this.out != null) {
                this.out.write(n);
            }
        }
        
        @Override
        public void write(byte[] b, int off, int len) throws IOException {
            this.count += len;
            if (this.out != null) {
                this.out.write(b, off, len);
            }
        }
        
        public void reset() {
            this.count = 0;
        }
        
        public int getSize() {
            return this.count;
        }
        
        @Override
        public String toString() {
            String s = super.toString();
            assert (s = this.stringForDebug()) != null;
            return s;
        }
        
        String stringForDebug() {
            return "<Sizer " + this.getSize() + ">";
        }
    }
}
